/* $Id: connection.cpp 411 2010-05-06 19:29:33Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "cfg.h"
using Cfg::ConnectionState;
#include "aprspacket.h"
#include "connection.h"

// std stuff.
#include <queue>
using std::queue;

// wx stuff.
#include <wx/thread.h>



Connection::Connection(ConnectionState const& state):
wxThread(wxTHREAD_JOINABLE),
state_(state), mutex_(), isStarted_(false), connected_(false),
inputSemaphore_(new wxSemaphore), outputSemaphore_(new wxSemaphore),
inputQueue_(new queue<AprsPacket*>), outputQueue_(new queue<AprsPacket const*>)
{
}

Connection::~Connection()
{
	// Delete semaphores
	delete inputSemaphore_;
	delete outputSemaphore_;

	// Delete everything in input and output queues and finally delete the queues.
	while (inputQueue_->size() > 0)
	{
		delete inputQueue_->front();
		inputQueue_->pop();
	}   

	delete inputQueue_;

	while (outputQueue_->size() > 0)
	{
		delete outputQueue_->front();
		outputQueue_->pop();
	}   

	delete outputQueue_;
}


// PUBLIC

bool Connection::isInitialized()
{
	// Check if we can read the situation.
	if ( mutex_.TryLock() == wxMUTEX_BUSY )
	{
		// No we can't, it's still initializing.
		return false;
	}
	
	// Yes we can, let's read.
	bool retval = isStarted_;
	mutex_.Unlock();
	
	return retval;
}

bool Connection::isConnected()
{
	mutex_.Lock();
	bool retval = connected_;
	mutex_.Unlock();
	
	return retval;
}

bool Connection::write(AprsPacket const* data)
{
	if (connected_) 
	{
		// Put the data to outputQueue to wait for sending.
		outputQueue_->push(data);
		
		// Tell the connection thread that there is data available.
		outputSemaphore_->Post();
		
		return true;
	}

	return false;
}

AprsPacket* Connection::readTimeout(unsigned int const& timeout)
{
	if (connected_)
	{
		// Wait until data is available or timeout occures.
		if (inputSemaphore_->WaitTimeout(timeout) == wxSEMA_NO_ERROR)
		{
			// For comments: see Connection::read()
			AprsPacket* input = inputQueue_->front();
			inputQueue_->pop();
			return input;
		}
	}
	
	// No connection or timeout.
	return NULL;
}

bool Connection::inputEnabled() const
{
	return state_ != Cfg::WRITEONLY;
}

bool Connection::outputEnabled() const
{
	return state_ != Cfg::READONLY;
}


// PROTECTED

void Connection::threadStarted()
{
	// This thread is now running, so it can lock it's mutex to hold main
	// thread at waitForConnect() until connection has been opened or
	// opening has failed.
	mutex_.Lock();
	
	// Tell main thread that it can begin to wait for our mutex to unlock.
	isStarted_ = true;
}

void Connection::gotConnected()
{
	// This thread is not running yet, when this function is called. There should
	// be no concurrency issues.
	connected_ = true;
}

void Connection::startupDone()
{
	// Unlock the mutex so that main thread can check our status.
	mutex_.Unlock();
}

void Connection::inputPush(AprsPacket* data)
{
	// Push data to input buffer...
	inputQueue_->push(data);

	// ...and Post() input semaphore to signal who ever is waiting that 
	// there data is available at the buffer.
	inputSemaphore_->Post();
}

AprsPacket const* Connection::outputPopWait(unsigned int const& timeout)
{
	wxSemaError result = outputSemaphore_->WaitTimeout(timeout);
	if (result == wxSEMA_NO_ERROR)
	{
		// Get the pointer from output buffer.
		AprsPacket const* data = outputQueue_->front();

		// Remove the element from the buffer.
		outputQueue_->pop();
		
		return data;
	}

	return NULL;
}

void Connection::gotDisconnected()
{
	mutex_.Lock();
	connected_ = false;
	mutex_.Unlock();
}
